/*	
  Copyright(c) 2011 Christoph Schmalz <christoph.schmalz@informatik.uni-erlangen.de>
  
  This file is licensed under the terms of of the GNU General Public License, 
  version 3, as published by the Free Software Foundation. You can find it at:
  http://www.gnu.org/licenses/gpl.html
*/

#include "stdafx.h"

#include "PhaseShiftPattern.h"
#include "Decoding.h"

#ifdef _DEBUG
	#pragma comment(lib, "opencv_core231d.lib")
	#pragma comment(lib, "opencv_imgproc231d.lib")
#else
	#pragma comment(lib, "opencv_core231.lib")
	#pragma comment(lib, "opencv_imgproc231.lib")
#endif
#pragma comment(lib, "zdll.lib") //for opencv ...

#include "imdebug.h"
void Imdebug(Mat& image, const char* desc);

void SetUpImages( const CPhaseShiftPattern& pattern, 
					imageArray_t& patternsGC, 
					imageArray_t& patternsPS,
					size_t nGrayCodeSlides,
					size_t nPhaseShiftSlides );

///////////////////////////////////////////////////////////////////////////////
// Function: main
// FunctionType: []
// Description: 
///////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
	//pattern resolution. depends on your display!
	const size_t resx = 800, resy = 600; //pixels
	const double pixelsize = 0.272; //mm
	
	//phase shift wavelength. typically 6 or 8. must be at least 6.
	const size_t wavelength = 8;
	//we use a 5-shift. it's very robust against gamma errors. need at least 3.
	const size_t nPhaseShiftBuckets = 5;

	//necessary number of gray code patterns depends on the pattern resolution. 
	//N slides can encode 2^N stripes. if eg 800 pixels are to be encoded and 
	//the stripe width is 4, we need 200 stripes, so 8 slides are enough. 
	//mind that the gray code stripe width is half the wavelength of the phase shift patterns.
	const size_t nGCSlides = 8; 

	//prepare the patterns
	//we need both vertical and horizontal stripes
	imageArray_t patternsGCv, patternsGCh, patternsPSv, patternsPSh;
	CPhaseShiftPattern pspV( resx, resy, wavelength, CPhaseShiftPattern::STRIPE_ORDER_LEFT_RIGHT, 1 );
	CPhaseShiftPattern pspH( resx, resy, wavelength, CPhaseShiftPattern::STRIPE_ORDER_TOP_DOWN, 1 );
	SetUpImages( pspV, patternsGCv, patternsPSv, nGCSlides, nPhaseShiftBuckets );
	SetUpImages( pspH, patternsGCh, patternsPSh, nGCSlides, nPhaseShiftBuckets );

	//now this would be the point where we capture the patterns with the camera to be calibrated
	//but for illustration we can use the patterns as they are
	//all the following functions support masks

	//decode gray code
	Mat QImgV(resy,resx,CV_8UC1), IndexImgV(resy,resx,CV_32SC1);
	Mat QImgH(resy,resx,CV_8UC1), IndexImgH(resy,resx,CV_32SC1);
	IndexImgV.setTo(0);
	IndexImgH.setTo(0);

	for (size_t i=0; i<nGCSlides; i++)
	{
		ClassifyGraycodeImg( patternsGCv[i*2], patternsGCv[i*2+1], 5, &QImgV, &IndexImgV, 1<<(nGCSlides-1-i) );
		ClassifyGraycodeImg( patternsGCh[i*2], patternsGCh[i*2+1], 5, &QImgH, &IndexImgH, 1<<(nGCSlides-1-i) );
	}

	//visualize
	//Imdebug(IndexImgH,"gc index");
	//Imdebug(QImgH,"gc quali");

	ConvertGrayCodeToBinary(&IndexImgV);
	ConvertGrayCodeToBinary(&IndexImgH);

	//decode phase shift
	//with the synthetic images we need to hack the overexposure threshold
	Mat phaseV,phaseH;
	EvaluatePhaseShift( patternsPSv, 10, &phaseV, &QImgV, 300 );
	EvaluatePhaseShift( patternsPSh, 10, &phaseH, &QImgH, 300 );

	//Imdebug(phaseV,"phaseV wrapped");
	//Imdebug(phaseH,"phaseH wrapped");
	//Imdebug(QImgV,"phase quali");

	//unwrap
	UnwrapPhaseWithGrayCode( &phaseV, &QImgV, &IndexImgV, pspV.GetNrOfGCStripes(), 10, true );
	UnwrapPhaseWithGrayCode( &phaseH, &QImgH, &IndexImgH, pspH.GetNrOfGCStripes(), 10, true );

	//Imdebug(phaseV,"phaseV unwrapped");
	//Imdebug(phaseH,"phaseH unwrapped");

	//make marks
	std::vector< CCalibMark* > marks;
	GenerateCalibPointsFromPhaseMap( &phaseV, &QImgV, &phaseH, &QImgH, 0.02f, 10, marks );

	//visualize marks
	Mat result(resy,resx,CV_8UC3);
	convertScaleAbs( phaseH, result, 255 );
	for ( std::vector< CCalibMark* >::iterator i=marks.begin(); i!=marks.end(); ++i)
		circle( result, Point( int((*i)->img.x+0.5), int((*i)->img.y+0.5) ), 2, Scalar(0,255,0,0) );
	
	//finally, the result
	Imdebug(result,"marks");

	//note that with these synthetic images, the mark image coords may be off by about 1e-4 pixels from their 
	//ideal locations (to compare, the final calibration errors are typically between 1e-2 and 1e-2 pixels).
	//this is due to quantization error in the pattern drawing. in real life this component is much smaller
	//than the errors due to camera noise and so on. experiment: change the parameters of the phase shift
	//(number of buckets, wavelength, gamma, quantization) and observe the influence.

	//convert mark phase coords to world coords
	TransformPhaseCoordsToWorldCoords( marks, pixelsize, resx, resy );

	//repeat for several poses
	//call calibration routine, eg calibrateCamera();
	//done!

	//cleanup
	for each (Mat* p in patternsGCv)
		delete p;
	for each (Mat* p in patternsGCh)
		delete p;
	for each (Mat* p in patternsPSv)
		delete p;
	for each (Mat* p in patternsPSh)
		delete p;

	return 0;
}


///////////////////////////////////////////////////////////////////////////////
// Function: SetUpImages
// FunctionType: []
// Description: 
//		helper function
///////////////////////////////////////////////////////////////////////////////
void SetUpImages( const CPhaseShiftPattern& pattern, 
					imageArray_t& patternsGC, 
					imageArray_t& patternsPS,
					size_t nGCSlides,
					size_t nPhaseShiftBuckets )
{
	Mat* pCurPattern;
	for (size_t i=0; i<nGCSlides; i++)
	{
		pCurPattern = new Mat();
		pattern.DrawGrayCode( *pCurPattern, i, nGCSlides, false );
		patternsGC.push_back( pCurPattern );

		pCurPattern = new Mat();
		pattern.DrawGrayCode( *pCurPattern, i, nGCSlides, true );
		patternsGC.push_back( pCurPattern );

		//visualize
		//Imdebug(*pCurPattern,"graycode");
	}

	for (size_t i=0; i<nPhaseShiftBuckets; i++)
	{
		pCurPattern = new Mat();
		pattern.DrawPhaseShiftPattern( *pCurPattern, i * M_2PI/nPhaseShiftBuckets );
		patternsPS.push_back( pCurPattern );

		//visualize
		//Imdebug(*pCurPattern,"phaseshift");
	}
}

///////////////////////////////////////////////////////////////////////////////
// Function: Imdebug
// FunctionType: []
// Description: 
//   wrapper for imdebug, a very convenient visual inspection tool
///////////////////////////////////////////////////////////////////////////////
void Imdebug(Mat& image, const char* desc)
{
	char* formatstring;
	switch (image.type())
	{
	case CV_8SC3:
	case CV_8UC3:
		formatstring = "bgr rgb=rgb w=%d h=%d b=8 t=%s %p";
		break;
	case CV_8UC4:
	case CV_8SC4:
	case CV_32SC1:
		formatstring = "bgra rgba=rgb_ w=%d h=%d b=8 t=%s %p"; 
		break;
	case CV_8SC1:
	case CV_8UC1:
		formatstring = "lum w=%d h=%d t=%s b=8 %p";
		break;
	case CV_32FC3:
		formatstring = "bgr w=%d h=%d b=32f t=%s %p";
		break;
	case CV_32FC1:
		formatstring = "lum w=%d h=%d t=%s b=32f %p";
		break;
	default:
		return;
	}

	imdebug( formatstring, image.size().width, image.size().height, desc, image.data );
}
