/*	
  Copyright(c) 2011 Christoph Schmalz <christoph.schmalz@informatik.uni-erlangen.de>
  
  This file is licensed under the terms of of the GNU General Public License, 
  version 3, as published by the Free Software Foundation. You can find it at:
  http://www.gnu.org/licenses/gpl.html
*/

#ifndef __DECODING_H__
#define __DECODING_H__

#include "util.h"

///////////////////////////////////////////////////////////////////////////////
// Function: EvaluatePhaseShift
// Description:
//	 phase shift decoding
///////////////////////////////////////////////////////////////////////////////

void EvaluatePhaseShift(
	const imageArray_t&		images,			//in: three or four images
	int						minDynamic,		//in: minimal dynamic to compute phase
	Mat*					pWrappedPhase,	//in: receives the wrapped phase
	Mat*					pMaskImg,		//in: receives mask
	int						iOverExpThresh=250);//in: threshold for overexposure


///////////////////////////////////////////////////////////////////////////////
// Function: UnwrapPhaseWithGrayCode
// Description:
//	unwrap phase with gray code stripe numbers.
//	phaseshift wavelength is assumed to be twice the gc stripe width
//	to solve cases of misalignment
///////////////////////////////////////////////////////////////////////////////

void UnwrapPhaseWithGrayCode(
 Mat*		pPhaseImg,				//in/out: phase image of type float
 Mat*		pQImg,					//in: quality image, may be NULL
 Mat*		pIndexImg,				//in: index image with integer stripe ids
 float		fNrOfGCStripes,		//in: total nr of gray code stripes
 uint8_t	ucMinQuality,		//in: minimal quality
 bool		bNormalizePhase);	//in: normalize unwrapped phase to [0:1]?


///////////////////////////////////////////////////////////////////////////////
// Function: ClassifyGraycodeImg
// FunctionType: []
// Description:
//   Classify all pixels of an image and set the current bit according to the
//   result of the classification.
//   It is assumed that all bits of the result image are 0 per default.
///////////////////////////////////////////////////////////////////////////////

bool // true if successful, false if an error occured.
 ClassifyGraycodeImg
  (Mat*	pPosImg,	//In: img taken under the Gray code pattern illu
   Mat*	pNegImg,	//In: img taken under inverse Gray code pattern illu
   int	iIntThres,	//In: Threshold specifying minimal intensity variation
   Mat*	pQImg,		//In: img for contrast (can be NULL)
   Mat*	pResultImg,	//In: img that receives result
   int	iCurBit);	//In: bit to set in the result image

///////////////////////////////////////////////////////////////////////////////
// Function: ConvertGrayCodeToBinary
// FunctionType: []
// Description:
//
///////////////////////////////////////////////////////////////////////////////
void ConvertGrayCodeToBinary(Mat* img);

///////////////////////////////////////////////////////////////////////////////
// Function: InterpolPhaseSubPix
// FunctionType: []
// Description:
///////////////////////////////////////////////////////////////////////////////
bool InterpolPhaseSubPix(
	const Mat* ximg, const Mat* ximg_m,			//phase x data
	const Mat* yimg, const Mat* yimg_m,			//phase y data
	int i, int j,								//base coords
	const std::vector<double> phaseX, 
	const std::vector<double> phaseY,			//phase values to find
	double minQuality,							//min mask value
	std::vector<CVec2D>& pointsXY,				//resulting coords
	double& intersectionAngle);

///////////////////////////////////////////////////////////////////////////////
// Function: FindPhaseCoords
// Description:
//	iterates over the images. where the quality is high enough and the 
//	phase changes over an integer multiple of the spacing in both x and y,
//	generate a new mark.
///////////////////////////////////////////////////////////////////////////////

void FindPhaseCoords( 
	const Mat* ximg, const Mat* ximg_m,
	const Mat* yimg, const Mat* yimg_m,
	const double dMinMarkQuality,
	const double dPhaseCoordSpacing,
	const double dPhaseCoordOffset,
	const float	 fMaxGradient,
	const double dMaxAngleDeviation,
	std::map<std::pair<size_t,size_t>,std::vector<CVec2D>>& detectedMarks );


///////////////////////////////////////////////////////////////////////////////
// Function: GenerateCalibPointsFromPhaseMap
// FunctionType: []
// Description: 
///////////////////////////////////////////////////////////////////////////////
bool GenerateCalibPointsFromPhaseMap(
		const Mat* ximg,			//phase x img
		const Mat* ximg_m,			//mask
		const Mat* yimg,			//phase y img
		const Mat* yimg_m,			//mask
		const float spacing,		//typically 0.02 (phase values are between 0 and 1)
		const uint8_t minQuality,
		std::vector<CCalibMark*>& calibPoints);	//phase + image location

///////////////////////////////////////////////////////////////////////////////
// Function: TransformPhaseCoordsToWorldCoords
// FunctionType: []
// Description: 
//	for cam calibration we observe the patterns on a display, ie
//	we know the image coords and need the world coords.
//	assuming a flat regular grid we can calculate
//	world coords from phase coords
///////////////////////////////////////////////////////////////////////////////

void TransformPhaseCoordsToWorldCoords(
			std::vector<CCalibMark*>& calibPoints,
			double dPixelSize,		//[mm]
			size_t iDeviceWidth,	//[pixels]
			size_t iDeviceHeight);	//[pixels]
#endif //__DECODING_H__