/*	
  Copyright(c) 2011 Christoph Schmalz <christoph.schmalz@informatik.uni-erlangen.de>
  
  This file is licensed under the terms of of the GNU General Public License, 
  version 3, as published by the Free Software Foundation. You can find it at:
  http://www.gnu.org/licenses/gpl.html
*/

#include "stdafx.h"
#include "util.h"
#include "decoding.h"

#include <algorithm>
#include <functional>
#include <numeric>

template <typename T>
void ComputeMeanAndStandardDeviation( const std::vector<T>& data, T& mean, T& sigma )
{
	switch (data.size())
	{
	case 0:
		mean = 0; sigma = 0; return;
	case 1:
		mean = data[0]; sigma = 0; return;
	default:
		mean = std::accumulate( data.begin(), data.end(), 0.0f )/ data.size();

		std::vector<T> zero_mean( data );
		std::transform( zero_mean.begin(), zero_mean.end(), zero_mean.begin(), std::bind2nd( std::minus<T>(), mean ) );

		sigma = std::inner_product( zero_mean.begin(), zero_mean.end(), zero_mean.begin(), 0 );
		sigma = sqrt( sigma / ( data.size() - 1 ) );
   }
}

///////////////////////////////////////////////////////////////////////////////
// Function: CLine2D::Intersect
///////////////////////////////////////////////////////////////////////////////

bool // true if the two lines, intersect, false if they are parallel
 CLine2D::Intersect
  (const CLine2D& o,	//in: other line
   CVec2D& r)			//out: intersection point
{
	double d = o.v.Cross(v);
    if (d == 0) 
		return false;

	double a = ( p.Cross(v) * o.v.x - o.p.Cross(o.v) * v.x ) / d;
	double b = ( p.Cross(v) * o.v.y - o.p.Cross(o.v) * v.y ) / d;

	r.Set(a,b);
   
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function: Intersect
// FunctionType: []
//   Intersects plane with ray (in 3D space)
///////////////////////////////////////////////////////////////////////////////

bool // true if the plane and the line intersect, false otherwise
 CPlane3D::Intersect(
	const CPlane3D&	other,
	CLine3D&		ray) const
{
	//line direction is perpendicular to both normal vectors
	ray.v = normal.Cross( other.normal );

	//normals parallel? no (unique) intersection
	if (ray.v.GetLengthSq()<1e-10)
		return false;

	//now we need a point on both planes
	//we have two equations but three variables
	//--> set one zero

	//we need to solve an equation system Mx=v
	Vec2d vector(-d,-other.d);

	//use largest component of v for best numerical stability
	if ( ( fabs(ray.v.z)>=fabs(ray.v.y) ) && ( fabs(ray.v.z)>=fabs(ray.v.x) ) )
	{
		//try z==0
		Matx22d matrix(normal.x, normal.y, other.normal.x, other.normal.y);
		Vec2d solution;

		solve( matrix, vector, solution, CV_SVD );
		ray.p.Set( solution(0), solution(1), 0 );
	}
	if ( ( fabs(ray.v.y)>=fabs(ray.v.z) ) && ( fabs(ray.v.y)>=fabs(ray.v.x) ) )
	{
		//try y==0
		Matx22d matrix(normal.x, normal.z, other.normal.x, other.normal.z);
		Vec2d solution;

		solve( matrix, vector, solution, CV_SVD );
		ray.p.Set( solution(0), 0, solution(1) );
	}
	if ( ( fabs(ray.v.x)>=fabs(ray.v.y) ) && ( fabs(ray.v.x)>=fabs(ray.v.z) ) )
	{
		//try x==0
		Matx22d matrix(normal.y, normal.z, other.normal.y, other.normal.z);
		Vec2d solution;

		solve( matrix, vector, solution, CV_SVD );
		ray.p.Set( 0, solution(0), solution(1) );
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function: FitPlaneLS
// FunctionType: []
// Description: 
//   Given an set of iSize data points adX, adY, and adZ this function computes
//   the plane of the form z=f(x,y) that gives the minimal least-square error.
///////////////////////////////////////////////////////////////////////////////

template <class T>
bool
 FitPlaneLS
 (const T		atX[],			//in: x-values of data points
  const T		atY[],			//in: y-values of data points
  const T		atZ[],			//in: z-values of data points
  const int		iNrVals,		//in: number of data points
  CPlane3D&		plane,			//in: receives plane
  const int		iStep=1)		//in: step to use when accessing the arrays
{
	if (iNrVals<3) 
		return false;

	Matx<T,3,3> A;
	Vec<T,3> b, x;

	A.zeros();
	b.all(0);

	for (int i=0, iIndex=0; i<iNrVals; i++, iIndex+=iStep)
	{
		A(0,0) += atX[iIndex]*atX[iIndex];
		A(0,1) += atX[iIndex]*atY[iIndex];
		A(0,2) += atX[iIndex];
		A(1,0) += atX[iIndex]*atY[iIndex];
		A(1,1) += atY[iIndex]*atY[iIndex];
		A(1,2) += atY[iIndex];
		A(2,0) += atX[iIndex];
		A(2,1) += atY[iIndex];
		A(2,2) += 1;

		b(0) += atX[iIndex]*atZ[iIndex];
		b(1) += atY[iIndex]*atZ[iIndex];
		b(2) += atZ[iIndex];
	}

	solve( A,b,x );

	//equation is ax+by+z+c=0
	plane = CPlane3D( CVec3D((double)x(0),(double)x(1),1).Normalize(), x(2));
	return true;
}


///////////////////////////////////////////////////////////////////////////////
// Function: InterpolPhaseSubPix
// FunctionType: []
// Description:
///////////////////////////////////////////////////////////////////////////////
bool InterpolPhaseSubPix(
	const Mat* ximg, const Mat* ximg_m,		//phase x data
	const Mat* yimg, const Mat* yimg_m,		//phase y data
	int i, int j,							//base coords
	const std::vector<double> phaseX, 
	const std::vector<double> phaseY,		//phase values to find
	double minQuality,						//min mask value
	std::vector<CVec2D>& pointsXY,			//resulting coords
	double& intersectionAngle)
{
	if (!ximg || !ximg_m || !yimg || !yimg_m)
		return false;

	//data
	const float* xdata = (float*)ximg->data;
	const float* ydata = (float*)yimg->data;
	//masks
	const uint8_t* maskx = ximg_m ? (uint8_t*)ximg_m->data : 0;
	const uint8_t* masky = yimg_m ? (uint8_t*)yimg_m->data : 0;
	//size
	int w = ximg->size().width;
	int h = ximg->size().height;

	//phase values (z) in x and y direction
	double zx[12], zy[12];

	//relative x and y coords of neigbors
	//first four are the "inner circle"
	double x[12] = { 0, 1, 0, 1, 0, 1, 2, 2, 1, 0, -1, -1 };
	double y[12] = { 0, 0, 1, 1, -1, -1, 0, 1, 2, 2, 1, 0 };

	//the sought after values define planes
	CPlane3D planeZX, planeZY;

	//local fit plane
	CPlane3D planeX, planeY;

	//intersection of z-planes with the local fit planes defines two lines
	CLine3D line3X,line3Y;

	//of which we need only the xy-part
	CLine2D line2X, line2Y;

	//get some more neighbors for the fit
	bool baddata = false;
	for (int k=0; k<12; k++)
	{
		//stop if we are too close to the border
		if (j+y[k] < 0 || i+x[k] < 0 || j+y[k] >= h || i+x[k] >= w)
		{
			baddata=true;
			break;
		}

		int curindex = int( (j+y[k])*w + (i+x[k]) );

		//we want data at all pixels
		if ( (maskx && maskx[curindex]<minQuality) || 
			 (masky && masky[curindex]<minQuality) )
		{
			baddata=true;
			break;
		}

		zx[k] = (double)xdata[ curindex ];
		zy[k] = (double)ydata[ curindex ];
	}

	if (baddata)
		return false;

	//fit planes ... todo: calc standard deviation
	if (!FitPlaneLS( x,y,zx,12,planeX ) || !FitPlaneLS( x,y,zy,12,planeY ))
		return false;

	//chance for debugging / plotting
	/*
	if (0)
	{
		AuxExportToMatlab( x, 12, 1, 1, "x" );
		AuxExportToMatlab( y, 12, 1, 1, "y" );
		AuxExportToMatlab( zx, 12, 1, 1, "zx" );
		AuxExportToMatlab( zy, 12, 1, 1, "zy" );
	}
	*/

	pointsXY.clear();
	for (size_t k=0; k<phaseX.size(); k++)
	{
		//intersect with the planes z=value
		planeZX.Set( CVec3D(0,0,1), phaseX[k] );
		planeZY.Set( CVec3D(0,0,1), phaseY[k] );	
		planeX.Intersect( planeZX, line3X );
		planeY.Intersect( planeZY, line3Y );

		//ignore the z-part now
		line2X.p = CVec2D( line3X.p.x,line3X.p.y );
		line2X.v = CVec2D( line3X.v.x,line3X.v.y ).Normalize();
		line2Y.p = CVec2D( line3Y.p.x,line3Y.p.y );
		line2Y.v = CVec2D( line3Y.v.x,line3Y.v.y ).Normalize();

		//intersect the two lines
		CVec2D ip;
		line2X.Intersect( line2Y, ip );

		//note that sometimes the intersection can be outside 
		//of the pixel itself, ie <0 or >1

		//transform from relative to absolute coords
		ip.x += i; ip.y += j;
		pointsXY.push_back(ip);
	}

	//the two lines should be approximately perpendicular to each other
	//calculate the intersection angle as a quality measure
	//(should be equal for all line pairs)
	intersectionAngle = acos( line2X.v.Normalize().Dot( line2Y.v.Normalize() ) );

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function: FindPhaseCoords
// Description:
//	iterates over the images. where the quality is high enough and the 
//	phase changes over an integer multiple of the spacing in both x and y,
//	generate a new mark.
///////////////////////////////////////////////////////////////////////////////

void FindPhaseCoords( 
	const Mat* ximg, const Mat* ximg_m,
	const Mat* yimg, const Mat* yimg_m,
	const double dMinMarkQuality,
	const double dPhaseCoordSpacing,
	const double dPhaseCoordOffset,
	const float	 fMaxGradient,
	const double dMaxAngleDeviation,
	std::map<std::pair<size_t,size_t>,std::vector<CVec2D>>& detectedMarks )
{
	if (!ximg || !ximg_m || !yimg || !yimg_m)
		return;

	//data
	const float* xdata = (float*)ximg->data;
	const float* ydata = (float*)yimg->data;
	//masks
	const uint8_t* maskx = ximg_m ? (uint8_t*)ximg_m->data : 0;
	const uint8_t* masky = yimg_m ? (uint8_t*)yimg_m->data : 0;
	//size
	int w = ximg->size().width;
	int h = ximg->size().height;

	//phase values (z) in x and y direction
	double zx[4], zy[4];

	//relative x and y coords of neigbors
	//first four are the "inner circle"
	double x[4] = { 0, 1, 0, 1 };
	double y[4] = { 0, 0, 1, 1 };

	//interface to phase interpol routine
	std::vector<CVec2D> pointXY;
	std::vector<double> phaseX, phaseY;

	for (int j=0; j<h-1; j++)
	{
		for (int i=0; i<w-1; i++)
		{
			//get phase values at the boundary of this pixel
			bool baddata = false;
			for (int k=0; k<4; k++)
			{
				int curindex = int( (j+y[k])*w + (i+x[k]) );

				//ignore masked pixels
				if ( (maskx && maskx[curindex]<dMinMarkQuality) || 
					 (masky && masky[curindex]<dMinMarkQuality) )
				{
					baddata = true;
					break;
				}

				//a bit inefficient, we could partially reuse the old data
				zx[k] = (double)xdata[ curindex ];
				zy[k] = (double)ydata[ curindex ];
			}

			if (baddata)
				continue;

			//we make no assumption whether the phase values are rising or falling
			double zxmax = MAX(MAX(MAX(zx[0],zx[1]),zx[2]),zx[3]);
			double zxmin = MIN(MIN(MIN(zx[0],zx[1]),zx[2]),zx[3]);
			double zymax = MAX(MAX(MAX(zy[0],zy[1]),zy[2]),zy[3]);
			double zymin = MIN(MIN(MIN(zy[0],zy[1]),zy[2]),zy[3]);

			if ( zxmax-zxmin>fMaxGradient || zymax-zymin>fMaxGradient )
				continue;

			size_t lox = size_t( (zxmin+dPhaseCoordOffset)/dPhaseCoordSpacing );
			size_t loy = size_t( (zymin+dPhaseCoordOffset)/dPhaseCoordSpacing );
			size_t hix = size_t( (zxmax+dPhaseCoordOffset)/dPhaseCoordSpacing );
			size_t hiy = size_t( (zymax+dPhaseCoordOffset)/dPhaseCoordSpacing );

			//check if the desired value is within this pixel
			if ( lox+1==hix && loy+1==hiy )
			{
				double angle;
				phaseX.clear();	phaseY.clear();
				phaseX.push_back(hix*dPhaseCoordSpacing-dPhaseCoordOffset);	
				phaseY.push_back(hiy*dPhaseCoordSpacing-dPhaseCoordOffset);	

				if ( InterpolPhaseSubPix( ximg, ximg_m, yimg, yimg_m,
											 i, j, 
											 phaseX, phaseY,
											 dMinMarkQuality,
											 pointXY, angle ) )
				{
					if ( fabs(90-R2D(angle))<dMaxAngleDeviation &&
						 pointXY[0].x >= 0 && pointXY[0].y >= 0 && 
						 pointXY[0].x < w && pointXY[0].y < h )
					{
						std::pair<size_t,size_t> markId( lox, loy );
						detectedMarks[ markId ].push_back( pointXY[0] );
					}
				}
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// Function: GenerateCalibPointsFromPhaseMap
// FunctionType: []
// Description: 
///////////////////////////////////////////////////////////////////////////////

bool // Result of the operation
GenerateCalibPointsFromPhaseMap(
		const Mat* ximg,		//phase x img
		const Mat* ximg_m,		//mask
		const Mat* yimg,		//phase y img
		const Mat* yimg_m,		//mask
		const float spacing,	//typically 0.02 (phase values are between 0 and 1)
		const uint8_t minQuality,
		std::vector<CCalibMark*>& calibPoints) //phase + image location
{
	//some basic sanity checks
	if (!ximg || !ximg_m || !yimg || !yimg_m)
		return false;

	//we do not want our phase coordinates to include 0 and 1 as the interval borders are unreliable in any case
	//instead we use an offset of half the spacing
	float offset = spacing/2;
 
	//now check where integer multiples of the phase coord spacing occur in the images
	std::map<std::pair<size_t,size_t>,std::vector<CVec2D>> detectedMarks;
	FindPhaseCoords( ximg,ximg_m,yimg,yimg_m,
		minQuality, spacing, offset,
		1e-2f,		//max phase gradient
		30,			//max angle deviation in degrees
		detectedMarks );

	//make a 5d array as expected by the calib routine
	calibPoints.clear();

	//we assume that the phase is normalized to one
	size_t marksPerRow = size_t( (1 - offset) / spacing );

	for (std::map<std::pair<size_t,size_t>,std::vector<CVec2D>>::iterator mark=detectedMarks.begin();
			mark!=detectedMarks.end(); ++mark)
	{
		if (mark->second.empty())
			continue;

		//convert the 2d mark id to a linear one (for identification)
		size_t ptNr = mark->first.first + marksPerRow * mark->first.second;

		//if the point has been detected uniquely
		if ( mark->second.size()==1 )
		{
			CVec2D pos = mark->second.front();

			uint8_t q = MIN( ximg_m->at<uint8_t>( int(pos.y+0.5), int(pos.x+0.5) ),
							 yimg_m->at<uint8_t>( int(pos.y+0.5), int(pos.x+0.5) ) );
			assert(q>=minQuality);

			calibPoints.push_back( 
				new CCalibMark( CVec3D( (mark->first.first+1)*spacing+offset, 
										(mark->first.second+1)*spacing+offset,
										0 ),
								CVec2D( pos.x, pos.y ), 
								q, ptNr ) );
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Function: TransformPhaseCoordsToWorldCoords
// FunctionType: []
// Description: 
//	for cam calibration we observe the patterns on a display, ie
//	we know the image coords and need the world coords.
//	assuming a flat regular grid we can calculate
//	world coords from phase coords
///////////////////////////////////////////////////////////////////////////////

void TransformPhaseCoordsToWorldCoords(
			std::vector<CCalibMark*>& calibPoints,
			double	dPixelSize,		//[mm]
			size_t	iDeviceWidth,	//[pixels]
			size_t	iDeviceHeight)	//[pixels]
{
	for (size_t j=0; j<calibPoints.size(); j++)
	{
		CCalibMark* p = calibPoints[j];
		p->world.x *= dPixelSize*iDeviceWidth;
		p->world.y *= dPixelSize*iDeviceHeight;
		p->world.z = 0;
	}
}
